# frozen_string_literal: false

require 'stringio'
require 'litbuild/blueprint_parser'
require 'litbuild/errors'
require 'litbuild/logfile_namer'
require 'litbuild/visitor'

module Litbuild
  class Blueprint
    class <<self
      # This should simply be the name of the directory where blueprints
      # of each specific type are expected to be found by Driver.
      def self.directory_name
        raise(SubclassResponsibility)
      end
    end

    attr_reader :active_phase
    attr_reader :base_grafs
    attr_writer :logfile_namer

    def self.descendants
      ObjectSpace.each_object(Class).select { |c| c < self }
    end

    # WARNING: in most cases, a freshly-created blueprint cannot be used
    # until it has also been _prepared_. (See below.)
    def initialize(text:)
      parser = BlueprintParser.new(text)
      @base_directives = parser.base_directives
      @phase_directives = parser.phase_directives
      @base_grafs = parser.base_grafs
      @phase_grafs = parser.phase_grafs
      @active_phase = @base_directives['default-phase']&.first
    end

    # This prepares a blueprint for use. This is separate from
    # initialize because parameters are not available until after all
    # blueprints have been parsed, and the logfile namer is not
    # available until after the LOGFILE_DIR parameter can be resolved.
    #
    # Parameters is a set of configuration parameters
    # Logfile_namer is used to generate log file names
    # Library is the BlueprintLibrary, used (for example) to find
    # dependencies.
    def prepare(parameters:, logfile_namer:, library:)
      @logfile_namer = logfile_namer
      @bp_library = library
      [@base_directives,
       @phase_directives,
       @base_grafs,
       @phase_grafs].each { |component| resolve_all(parameters, component) }
    end

    # Dependencies need to be handled before their dependants; that's
    # handled here. Subclasses should run `super` before doing anything
    # else with the Visitor, or call the send_to_dependencies method
    # directly, so that this happens properly!
    def accept(visitor:)
      send_to_dependencies(visitor: visitor)
    end

    # Return a copy of this blueprint that is initialized to operate on
    # the specified phase.
    def for_phase(new_active_phase)
      return self unless phases?

      phased_blueprint = clone
      phased_blueprint.phase = new_active_phase if new_active_phase
      phased_blueprint
    end

    def phases
      @phase_directives.keys
    end

    def phases?
      !@phase_directives.empty?
    end

    def phase_grafs
      @phase_grafs[active_phase]
    end

    def directives
      if active_phase
        @phase_directives[active_phase]
      else
        @base_directives
      end
    end

    def [](directive)
      directives[directive]
    end

    # This is just like the [] operator method except that it always
    # returns only the first value for a directive -- which is helpful
    # for all the directives that are only ever supposed to have a
    # single value, like `name` or `full-name`.
    def value(directive)
      self[directive]&.first
    end

    def name
      value('name')
    end

    def name_with_phase
      if active_phase
        "#{name}::#{active_phase}"
      else
        name
      end
    end

    def full_name
      value('full-name')
    end

    def file_name
      if active_phase
        "#{name}-#{active_phase}"
      else
        name
      end.tr(' ', '-')
    end

    def header_text
      value('full-name')
    end

    def header_text_with_phase
      "#{header_text} (#{active_phase} phase)"
    end

    def logfile(stage_name, phase_name = nil)
      phase = phase_name&.tr(' ', '_')
      @logfile_namer.path_for(name, phase, stage_name)
    end

    def parameter_defaults
      values = {}
      all_directive_sets = [@base_directives, @phase_directives.values].flatten
      all_directive_sets.each do |dirset|
        next unless dirset.key?('parameter')

        dirset['parameter'].each do |a_param|
          val = a_param['default'].first
          values[a_param['name'].first] = if val == '(empty)'
                                            ''
                                          else
                                            val
                                          end
        end
      end
      values
    end

    def target_name
      if active_phase
        "#{name}::#{active_phase}"
      else
        name
      end
    end

    def success_line
      'SUCCESS'
    end

    def failure_line
      'FAILURE'
    end

    # If a dependency is declared both without a phase (typically in the
    # base directives for the blueprint) and with a phase (typically in
    # a phase of that blueprint), throw away the phaseless declaration
    # to avoid including two versions of the dependency.
    def deduped_dependency_names
      dep_names = self['depends-on'].clone || []
      deps_with_phase = dep_names.select { |dep| dep.match?(/::/) }
      deps_with_phase.each do |dep|
        dep_without_phase = dep.split('::')[0]
        dep_names.delete(dep_without_phase)
      end
      dep_names
    end

    protected

    def send_to_dependencies(visitor:)
      @bp_library.dependencies_for(self).each do |bp|
        bp.accept(visitor: visitor)
      end
    end

    def phase=(phase)
      bptype = self.class.name.split('::').last
      msg = "#{bptype} #{name} does not have a phase '#{phase}'"
      raise(InvalidParameter, msg) unless phase.nil? || phases.include?(phase)

      @active_phase = phase
    end

    private

    def resolve_all(parameters, something)
      case something
      when Hash
        something.values.each { |element| resolve_all(parameters, element) }
      when Array
        something.each { |element| resolve_all(parameters, element) }
      when String
        resolve(parameters: parameters, a_string: something)
      end
    end

    def resolve(parameters:, a_string:)
      params = a_string.scan(/PARAM\[([A-Z_]+)\]/).flatten.sort.uniq
      params.each do |p|
        unless parameters[p]
          raise(ParameterMissing, "Parameter #{p} is not defined")
        end

        a_string.gsub!(/PARAM\[#{p}\]/, parameters[p])
      end
      a_string
    end
  end
end
