# frozen_string_literal: true

require 'litbuild/blueprint_library'
require 'litbuild/ascii_doc_visitor'
require 'litbuild/bash_script_visitor'
require 'litbuild/source_files_visitor'
require 'litbuild/url_visitor'

module Litbuild
  # This is what the command-line program `lb` uses to do all the work,
  # aside from parsing the command line. It initializes a blueprint
  # library and provides methods for exercising all the functionality of
  # litbuild: writing bash scripts, writing documents, querying package
  # details, etc.
  #
  # All the actual work is done by Visitors that are handed to the
  # top-level blueprint target (typically specified on the command line).
  class Driver
    REQUIRED_PARAMS = %w[DOCUMENT_DIR LOGFILE_DIR PATCH_DIR
                         SCRIPT_DIR TARFILE_DIR WORK_SITE].freeze

    def initialize(logfile_namer_class: Litbuild::LogfileNamer)
      @bplib = BlueprintLibrary.new(logfile_namer_class: logfile_namer_class)
    end

    def library
      @bplib
    end

    def params
      @bplib.parameters
    end

    def download_urls_for(target:)
      uv = UrlVisitor.new
      dispatch(visitor: uv, target: target)
      uv.urls
    end

    def source_files_for(target:)
      sfv = SourceFilesVisitor.new
      dispatch(visitor: sfv, target: target)
      sfv.files
    end

    def write_scripts_for(target:)
      bsv = BashScriptVisitor.new(parameters: params)
      dispatch(visitor: bsv, target: target)
      bsv.write_sudoers
      bsv.write_toplevel_script(target: target)
    end

    def write_document_for(target:)
      adv = AsciiDocVisitor.new(parameters: params)
      bp = library.blueprint_for(target: target)
      bp.accept(visitor: adv)
      adv.write_toplevel_doc(blueprint: bp)
    end

    private

    def dispatch(visitor:, target:)
      library.blueprint_for(target: target).accept(visitor: visitor)
    end
  end
end
