# frozen_string_literal: true

require 'litbuild/blueprint'
require 'litbuild/errors'
require 'litbuild/multi_part_visitor'

module Litbuild
  class Section < Blueprint
    def self.directory_name
      'sections'
    end

    def accept(visitor:)
      super
      visitor.in_subdirectory(name) do
        components.each do |bp|
          bp.accept(visitor: visitor)
        end
      end
      visitor.visit_section(section: self)
      return unless visitor.is_a?(MultiPartVisitor)

      self['other-parts']&.each do |part|
        part_bp = @bp_library.blueprint_for(target: part)
        visitor.visit_part(part_bp)
      end
      self['appendices']&.each do |app|
        app_bp = @bp_library.blueprint_for(target: app)
        visitor.visit_appendix(app_bp)
      end
    end

    def components
      return @components if @components

      blueprints = self['blueprints'].clone || []
      if $DEBUG
        warn("Explicit blueprints for section #{name}:" \
             " #{blueprints.join(', ')}")
      end
      blueprints << automatic_inclusions
      deduped = blueprints.flatten.uniq
      @components = deduped.map do |bp|
        @bp_library.blueprint_for(target: bp)
      end
      @components
    end

    def automatic_inclusions
      auto_adds = []
      @bp_library.blueprints.each_value do |bp|
        if bp.phases.include?(name)
          auto_adds << "#{bp.name}::#{name}"
        elsif bp['in-section']&.include?(name)
          auto_adds << bp.name
        end
      end
      if $DEBUG
        warn("Automatically-added blueprints for section #{name}:" \
             " #{auto_adds.sort.join(', ')}")
      end
      auto_adds.sort
    end

    def success_line
      'HUGE SUCCESS'
    end

    def failure_line
      'DISMAL FAILURE'
    end
  end
end
